/*
    Romfs file system driver for Windows.
    Copyright (C) 1999-2015 Bo Brantn.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "ntifs.h"
#include "fsd.h"
#include "rom_fs.h"
#include "border.h"

#pragma code_seg(FSD_PAGED_CODE)

NTSTATUS
FsdIsDeviceRomfs (
    IN PDEVICE_OBJECT DeviceObject
    )
{
    PUCHAR          Buffer;
    LARGE_INTEGER   Offset;
    NTSTATUS        Status;

    PAGED_CODE();

    ASSERT(DeviceObject != NULL);

    Buffer = FsdAllocatePool(NonPagedPool, SECTOR_SIZE, '1ceR');

    if (!Buffer)
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    Offset.QuadPart = ROMFS_MAGIC_OFFSET;

    Status = FsdReadBlockDeviceOverrideVerify(
        DeviceObject,
        &Offset,
        SECTOR_SIZE,
        Buffer
        );

    if (!NT_SUCCESS(Status) ||
        RtlCompareMemory(Buffer, ROMFS_MAGIC, sizeof(ROMFS_MAGIC) - sizeof(UCHAR)) != (sizeof(ROMFS_MAGIC) - sizeof(UCHAR))
        )
    {
        Status = STATUS_UNRECOGNIZED_VOLUME;
    }

    FsdFreePool(Buffer);

    return Status;
}

NTSTATUS
FsdIsDeviceSameRomfs (
    IN PDEVICE_OBJECT   DeviceObject,
    IN ULONG            CheckSum
    )
{
    struct romfs_super_block*   romfs_super_block;
    LARGE_INTEGER               Offset;
    NTSTATUS                    Status;

    PAGED_CODE();

    ASSERT(DeviceObject != NULL);

    romfs_super_block = FsdAllocatePool(NonPagedPool, SECTOR_SIZE, '2ceR');

    if (!romfs_super_block)
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    Offset.QuadPart = ROMFS_MAGIC_OFFSET;

    Status = FsdReadBlockDeviceOverrideVerify(
        DeviceObject,
        &Offset,
        SECTOR_SIZE,
        romfs_super_block
        );

    if (!NT_SUCCESS(Status) ||
        RtlCompareMemory(romfs_super_block, ROMFS_MAGIC, sizeof(ROMFS_MAGIC) - sizeof(UCHAR)) != (sizeof(ROMFS_MAGIC) - sizeof(UCHAR)) ||
        be32_to_cpu(romfs_super_block->checksum) != CheckSum
        )
    {
        Status = STATUS_WRONG_VOLUME;
    }

    FsdFreePool(romfs_super_block);

    return Status;
}

NTSTATUS
FsdReadInodeByIndex (
    IN PDEVICE_OBJECT           DeviceObject,
    IN ULONG                    Index,
    IN OUT struct romfs_inode*  Inode
    )
{
    PUCHAR          Buffer;
    LARGE_INTEGER   Offset;
    NTSTATUS        Status;
    PDEVICE_OBJECT  DeviceToVerify;

    PAGED_CODE();

    ASSERT(DeviceObject != NULL);
    ASSERT(Inode != NULL);

    Buffer = (PUCHAR) FsdAllocatePool(NonPagedPool, 1024, '2mTR');

    if (Buffer == NULL)
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    // Byte offset to the inode
    Index = be32_to_cpu(Index) & ROMFH_MASK;

    //KdPrint((DRIVER_NAME ": FsdReadInodeByIndex: Index: %#x\n", Index));

    // Sector align the offset
    Offset.QuadPart = Index & ~(SECTOR_SIZE - 1);

    Status = FsdReadBlockDevice(
        DeviceObject,
        &Offset,
        1024,
        Buffer
        );

    if (Status == STATUS_VERIFY_REQUIRED)
    {
        DeviceToVerify = IoGetDeviceToVerify(PsGetCurrentThread());

        IoSetDeviceToVerify(PsGetCurrentThread(), NULL);

        Status = IoVerifyVolume(DeviceToVerify, FALSE);

        if (NT_SUCCESS(Status))
        {
            Status = FsdReadBlockDevice(
                DeviceObject,
                &Offset,
                1024,
                Buffer
                );
        }
    }

    RtlCopyMemory(
        Inode,
        Buffer + (Index & (SECTOR_SIZE - 1)),
        sizeof(struct romfs_inode) + ROMFS_MAXFN
        );

    FsdFreePool(Buffer);

    return Status;
}

#pragma code_seg() // end FSD_PAGED_CODE

NTSTATUS
FsdReadFileData (
    IN PDEVICE_OBJECT       DeviceObject,
    IN ULONG                Index,
    IN struct romfs_inode*  Inode,
    IN PLARGE_INTEGER       Offset,
    IN ULONG                Length,
    IN OUT PVOID            Buffer
    )
{
    LARGE_INTEGER   PhysicalOffset;
    ULONG           PhysicalLength;
    PUCHAR          PhysicalBuffer;
    NTSTATUS        Status;

    ASSERT(DeviceObject != NULL);
    ASSERT(Inode != NULL);
    ASSERT(Offset != NULL);
    ASSERT(Buffer != NULL);

    KdPrint((
        DRIVER_NAME
        ": FsdReadFileData: Index: %#x Offset: %I64u Length: %u\n",
        Index,
        Offset->QuadPart,
        Length
        ));

    //
    // On Romfs file data is 16 byte aligned and continuous while read/write
    // operations on storage devices must be sector aligned and have an length
    // that is a multiple of the sector size. If the file data happens to be
    // sector aligned we just calculate the physical offset to it and do a read
    // from the storage device. If the file data is not sector aligned we
    // allocate a temporary buffer, do a sector aligned read to it and copy the
    // right data from the temporary buffer to the callers buffer.
    //

    if (!((Index + SIZEOF_ROMFS_INODE(Inode) + Offset->QuadPart) &
        (SECTOR_SIZE - 1))
        )
    {
        PhysicalOffset.QuadPart =
            Index + SIZEOF_ROMFS_INODE(Inode) + Offset->QuadPart;

        Status = FsdReadBlockDevice(
            DeviceObject,
            &PhysicalOffset,
            Length,
            Buffer
            );
    }
    else
    {
        PhysicalOffset.QuadPart =
            (Index + SIZEOF_ROMFS_INODE(Inode) + Offset->QuadPart) &
            ~(SECTOR_SIZE - 1);

        PhysicalLength = Length + SECTOR_SIZE;

        PhysicalBuffer = (PUCHAR) FsdAllocatePool(
            NonPagedPool,
            PhysicalLength,
            '3mTR'
            );

        if (PhysicalBuffer == NULL)
        {
            return STATUS_INSUFFICIENT_RESOURCES;
        }

        Status = FsdReadBlockDevice(
            DeviceObject,
            &PhysicalOffset,
            PhysicalLength,
            PhysicalBuffer
            );

        RtlCopyMemory(
            Buffer,
            PhysicalBuffer +
            ((Index + SIZEOF_ROMFS_INODE(Inode) + Offset->QuadPart) &
            (SECTOR_SIZE - 1)),
            Length
            );

        FsdFreePool(PhysicalBuffer);
    }

    return Status;
}
